#include "floatimage.h"
#include <stdlib.h>

//**************************************************************************
// Global variables:
//**************************************************************************

float gamma_table[256];
float cur_gamma = 0.0;


//**************************************************************************
// Some utility functions we need
//**************************************************************************
int ReadBMPHeader(FILE *infile, BITMAPINFOHEADER *i)
{
  BITMAPFILEHEADER head;
	
  if ((infile == NULL) || (i == NULL)) return 0;

  if (fread(&head, sizeof(BITMAPFILEHEADER), 1, infile) != 1) return 0;
  if (fread(i, sizeof(BITMAPINFOHEADER), 1, infile) != 1) return 0;

  fseek(infile, head.bfOffBits, SEEK_SET);
  
  return 1;
}


int WriteBMPHeader(BITMAPINFOHEADER *i, FILE *outfile)
{
	BITMAPFILEHEADER bmfh;

	bmfh.bfType = *((WORD *) "BM");
	bmfh.bfOffBits = sizeof(BITMAPFILEHEADER) + sizeof(BITMAPINFOHEADER);
	if (i->biBitCount == 8)
		bmfh.bfOffBits += 256 * sizeof(RGBQUAD);
	bmfh.bfSize = bmfh.bfOffBits + i->biWidth * i->biHeight;
	bmfh.bfReserved1 = 0;
	bmfh.bfReserved2 = 0;

	if (fwrite(&bmfh, sizeof(BITMAPFILEHEADER), 1, outfile) != 1) return 0;
	if (fwrite(i, sizeof(BITMAPINFOHEADER), 1, outfile) != 1) return 0;

	if (i->biBitCount == 8)
	{
		if (fwrite(&i[1], 256 * sizeof(RGBQUAD), 1, outfile) != 1) return 0;
	}

	return 1;
}


void CreateGammaTable(float gamma, float *table)
{
	for (int x = 0; x < 256; x++)
		table[x] = (float) pow((double)x / 255.0, (double)gamma);
}


void InitGamma(float gamma)
{
	if (cur_gamma == gamma)
		return;

	CreateGammaTable(gamma, gamma_table);

	cur_gamma = gamma;
}


int RecursiveGCD(int a, int b)				// should make this non-recursive at some point
{
//	assert(a<b);
	
	if (a == 0)								// keep calculating until we get GCD(0,r)
		return abs(b);						// GCD > 0
	else
		return RecursiveGCD(b % a, a);		// Calculating GCD(r,a)
}



#define REVERSE2BYTES(source, dest)		\
{										\
	char *j = (char *) source;			\
	char *dj = (char *) dest;			\
	dj[0] = j[1];						\
	dj[1] = j[0];						\
}


#define REVERSEBYTES(source, dest)		\
{										\
	char *j = (char *) source;			\
	char *dj = (char *) dest;			\
	dj[0] = j[3];						\
	dj[1] = j[2];						\
	dj[2] = j[1];						\
	dj[3] = j[0];						\
}


#define REVERSE2BYTES_INPLACE(sourcedest)		\
{												\
	char *j = (char *) sourcedest;				\
	j[0] ^= j[1]; j[1] ^= j[0]; j[0] ^= j[1];	\
}


#define REVERSEBYTES_INPLACE(sourcedest)		\
{												\
	char *j = (char *) sourcedest;				\
	j[0] ^= j[3]; j[3] ^= j[0]; j[0] ^= j[3];	\
	j[1] ^= j[2]; j[2] ^= j[1]; j[1] ^= j[2];	\
}


#ifndef M_PI
#define M_PI	3.14159265358979323846264338327950288
#endif



// Floating point format:
//
//			SEEEEEEEEMMMMMMMMMMMMMMMMMMMMMMM
//			|       |       |       |      |
//			0       8       16      24     31
//
//
//		value = (-1)^(S) * (1.MMMMMMMMMMMMMMMMMMMMMMM) * 2^(signed binary EEEEEEEE)
//

inline unsigned short int float2short(float f)
{
  unsigned int bits = * (unsigned int *) &f;
  return (bits >> 15);
}


inline unsigned short int signedfloat2short(float f)
{
  unsigned int bits = * (unsigned int *) &f;
  return (bits >> 16);
}


inline float short2float(unsigned short int s)
{
	unsigned int bits = s << 15;
	return * (float *) &bits;
}

inline float short2signedfloat(unsigned short int s)
{
	unsigned int bits = s << 16;
	return * (float *) &bits;
}


void floatimage::ClipLower(float min, float replace)
{
	float *f = &line[0][0].r;
	float *last = &line[sizey-1][sizex-1].b;

	while (f <= last)
	{
		float r = f[0];
		float g = f[1];
		float b = f[2];

		if ((f[0] < min) && (f[1] < min) && (f[2] < min))
		{
			f[0] = replace;
			f[1] = replace;
			f[2] = replace;
		}

		f += 3;
	}
}



floatimage&	floatimage::operator = (const floatimage &in) 
{
	float *f = &in.line[0][0].r;
	float *last = &in.line[in.sizey-1][in.sizex-1].b;
	float *fout;

	Allocate(in.sizex,in.sizey);
	fout = &line[0][0].r;
	
	while (f<=last)
	{
		*fout++ = *f++;
		*fout++ = *f++;
		*fout++ = *f++;
	}

	return *this;
}


floatimage::floatimage(const floatimage &in) 
{
	line = NULL;
	float *f = &in.line[0][0].r;
	float *last = &in.line[in.sizey-1][in.sizex-1].b;
	float *fout;

	Allocate(in.sizex,in.sizey);
	fout = &line[0][0].r;
	
	while (f<=last)
	{
		*fout++ = *f++;
		*fout++ = *f++;
		*fout++ = *f++;
	}
}



void floatimage::Allocate(int x, int y)
{
	if ((sizex == x) && (sizey == y))
		return;						// already a good size!  Yay!

	Release();

	if ((x > 0) && (y > 0))
	{
		line = (fRGB **) malloc(sizeof(fRGB *) * y);

		if (line)
		{
			line[0] = (fRGB *) malloc(sizeof(fRGB) * x * y);

			if (!line[0])
				Release();
			else
				for (int cy = 1; cy < y; cy++)
					line[cy] = &line[0][cy * x];

			sizex = x;
			sizey = y;
		}
	}
}



void floatimage::Release()
{
	if (line)
	{
		if (line[0])
		{
			free(line[0]);
			line[0] = NULL;
		}
		free(line);
		line = NULL;
	}
	sizex = 0;
	sizey = 0;
}



void floatimage::SwapByteOrder()
{
	for (int y = 0; y < sizey; y++)
		for (int x = 0; x < sizex; x++)
		{
			REVERSEBYTES_INPLACE(&line[y][x].r);
			REVERSEBYTES_INPLACE(&line[y][x].g);
			REVERSEBYTES_INPLACE(&line[y][x].b);
		}
}



int floatimage::LoadPFM(const char *filename)
{
	FILE *infile = fopen(filename, "rb");

	int a = fgetc(infile);
	int b = fgetc(infile);
	int junk = fgetc(infile);

	if ((a != 'P') || ((b != 'F') && (b != 'f')))
	{
		fclose(infile);
		return 0;
	}

	b = (b == 'F');		// 'F' = RGB,  'f' = monochrome

	int x, y;
	fscanf(infile, "%d %d%c", &x, &y, &junk);
	if ((x <= 0) || (y <= 0))
	{
		fclose(infile);
		return 0;
	}

	float scalefactor;
	fscanf(infile, "%f%c", &scalefactor, &junk);


	Allocate(x, y);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	a = sizex * (b ? 3 : 1);
	float *fbuf = new float[a];
	fRGB *cur = &line[0][0];
	for (y = 0; y < sizey; y++)
	{
		if (fread(fbuf, sizeof(float), a, infile) != (size_t) a)
		{
			fclose(infile);
			delete fbuf;
			return 0;
		}

		float *temp = fbuf;
		for (x = 0; x < sizex; x++)
		{
			if (b)			// color
			{
				if (scalefactor > 0.0)		// MSB
				{
					REVERSEBYTES(temp++, &cur->r);
					REVERSEBYTES(temp++, &cur->g);
					REVERSEBYTES(temp++, &cur->b);
				}
				else						// LSB
				{
					cur->r = *temp++;
					cur->g = *temp++;
					cur->b = *temp++;
				}
			}
			else			// black and white
			{
				float c;
				if (scalefactor > 0.0)		// MSB
				{
					REVERSEBYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = c;
			}
			cur++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	return 1;
}


int floatimage::SavePFM(const char *filename)
{
	FILE *outfile = fopen(filename, "wb");

	fputc('P', outfile);
	fputc('F', outfile);
	fputc(0x0a, outfile);

	fprintf(outfile, "%d %d", sizex, sizey);
	fputc(0x0a, outfile);

	fprintf(outfile, "%f", -1.0f);
	fputc(0x0a, outfile);

	for (int y = sizey - 1; y >= 0; y--)
	{
		fwrite(&line[y][0].r, sizeof(float), 3 * sizex, outfile);
	}

	fclose(outfile);

	return 1;
}


int floatimage::SaveRAW(const char *filename)
{
	FILE *outfile = fopen(filename, "wb");
	if (!outfile)
	{
		return 0;
	}

	for (int y = sizey - 1; y >= 0; y--)
	{
		fwrite(&line[y][0].r, sizeof(float), 3 * sizex, outfile);
	}

	fclose(outfile);
	return 1;
}



int floatimage::LoadRAW_Bytes(const char *filename, int width, int height, int rgb, float gamma, int header)
{
	InitGamma(gamma);
	
	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}

	fseek(infile, header, SEEK_CUR);

	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int numbytes = (rgb ? 3 * width : width);
	unsigned char *fbuf = new unsigned char[numbytes];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(unsigned char), numbytes, infile) != (size_t) numbytes)
		{
			delete fbuf;
			fclose(infile);
			return 0;
		}

		unsigned char *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				cur->r = gamma_table[*temp++];
				cur->g = gamma_table[*temp++];
				cur->b = gamma_table[*temp++];
			}
			else
			{
				cur->r = cur->g = cur->b = gamma_table[*temp++];
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();
	return 1;
}


int floatimage::LoadRAW_Shorts(const char *filename, int width, int height, int rgb, int byte_reverse, int header)
{
	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}

	fseek(infile, header, SEEK_CUR);
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int numshorts = (rgb ? 3 * width : width);
	short int *fbuf = new short int[numshorts];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(short int), numshorts, infile) != (size_t) numshorts)
		{
			delete fbuf;
			fclose(infile);
			return 0;
		}

		short int *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				if (byte_reverse)		// MSB
				{
					REVERSE2BYTES(temp++, &cur->r);
					REVERSE2BYTES(temp++, &cur->g);
					REVERSE2BYTES(temp++, &cur->b);
				}
				else
				{
					cur->r = *temp++;
					cur->g = *temp++;
					cur->b = *temp++;
				}
			}
			else
			{
				short c;
				if (byte_reverse)		// MSB
				{
					REVERSE2BYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = c;
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	return 1;
}


int floatimage::LoadRAW_UnsignedShorts(const char *filename, int width, int height, int rgb, int byte_reverse, int header)
{
	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}
	
	fseek(infile, header, SEEK_CUR);

	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int numshorts = (rgb ? 3 * width : width);
	unsigned short int *fbuf = new unsigned short int[numshorts];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(unsigned short int), numshorts, infile) != (size_t) numshorts)
		{
			delete fbuf;
			fclose(infile);
			return 0;
		}

		unsigned short int *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				if (byte_reverse)		// MSB
				{
					REVERSE2BYTES(temp++, &cur->r);
					REVERSE2BYTES(temp++, &cur->g);
					REVERSE2BYTES(temp++, &cur->b);
				}
				else
				{
					cur->r = *temp++;
					cur->g = *temp++;
					cur->b = *temp++;
				}
			}
			else
			{
				unsigned short c;
				if (byte_reverse)		// MSB
				{
					REVERSE2BYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = c;
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	return 1;
}


int floatimage::LoadRAW_Longs(const char *filename, int width, int height, int rgb, int byte_reverse, int header)
{
	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}

	fseek(infile, header, SEEK_CUR);
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int numlongs = (rgb ? 3 * width : width);
	int *fbuf = new int[numlongs];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(int), numlongs, infile) != (size_t) numlongs)
		{
			delete fbuf;
			fclose(infile);
			return 0;
		}

		int *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &cur->r);
					REVERSEBYTES(temp++, &cur->g);
					REVERSEBYTES(temp++, &cur->b);
				}
				else
				{
					cur->r = (float) (*temp++);
					cur->g = (float) (*temp++);
					cur->b = (float) (*temp++);
				}
			}
			else
			{
				unsigned long c;
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = (float) c;
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	return 1;
}


int floatimage::LoadRAW_UnsignedLongs(const char *filename, int width, int height, int rgb, int byte_reverse, int header)
{
	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}

	fseek(infile, header, SEEK_CUR);
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int numlongs = (rgb ? 3 * width : width);
	unsigned int *fbuf = new unsigned int[numlongs];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(unsigned int), numlongs, infile) != (size_t) numlongs)
		{
			delete fbuf;
			fclose(infile);
			return 0;
		}

		unsigned int *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &cur->r);
					REVERSEBYTES(temp++, &cur->g);
					REVERSEBYTES(temp++, &cur->b);
				}
				else
				{
					cur->r = (float) (*temp++);
					cur->g = (float) (*temp++);
					cur->b = (float) (*temp++);
				}
			}
			else
			{
				unsigned long c;
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = (float) c;
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	return 1;
}


int floatimage::LoadRAW(const char *filename, int width, int height, int rgb, int byte_reverse, int header)
{
	FILE *infile = fopen(filename, "rb");
	if (!infile)
	{
		return 0;
	}

	fseek(infile, header, SEEK_CUR);
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int numfloats = (rgb ? 3 * width : width);
	float *fbuf = new float[numfloats];

	fRGB *cur = &line[0][0];
	for (int y = 0; y < height; y++)
	{
		if (fread(fbuf, sizeof(float), numfloats, infile) != (size_t) numfloats)
		{
			delete fbuf;
			fclose(infile);
			return 0;
		}

		float *temp = fbuf;
		for (int x = 0; x < width; x++)
		{
			if (rgb)
			{
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &cur->r);
					REVERSEBYTES(temp++, &cur->g);
					REVERSEBYTES(temp++, &cur->b);
				}
				else						// LSB
				{
					cur->r = *temp++;
					cur->g = *temp++;
					cur->b = *temp++;
				}
			}
			else
			{
				float c;
				if (byte_reverse)		// MSB
				{
					REVERSEBYTES(temp++, &c);
				}
				else						// LSB
					c = *temp++;
				cur->r = cur->g = cur->b = c;
			}

			cur ++;
		}
	}

	delete fbuf;
	fclose(infile);

	FlipVertical();

	return 1;
}



int floatimage::LoadBMP_SubImage(int startx, int starty, int width, int height, const char *filename, float image_gamma)
{
	FILE *infile = fopen(filename, "rb");

	if (infile == NULL)
		return 0;

	BITMAPINFOHEADER info;
	if (!ReadBMPHeader(infile, &info))
	{
		fclose(infile);
		return 0;
	}
	
	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int stride = sizeof(RGB) * info.biWidth;			// stride between lines in bytes
	stride = (stride + 3) & (~0x03);					// rounded up to nearest LONG boundary

	InitGamma(image_gamma);

	RGB *buf = new RGB[width];

	fseek(infile, stride * starty, SEEK_CUR);		// read in all the junk before the area we want

	unsigned int ofs = ftell(infile);
	for (int y = 0; y < height; y++)
	{
		fseek(infile, ofs + startx * sizeof(RGB), SEEK_SET);	// jump to start of data

		if (fread(buf, width * sizeof(RGB), 1, infile) != 1)	// read relevant data
		{
			Release();
			delete [] buf;
			fclose(infile);
			return 0;
		}

		RGB *cur = buf;											// put data into the floatimage
		fRGB *curline = line[sizey - 1 - y];
		for (int x = 0; x < width; x++)
		{
			curline[x].r = gamma_table[cur->rgbRed];
			curline[x].g = gamma_table[cur->rgbGreen];
			curline[x].b = gamma_table[cur->rgbBlue];
			cur++;
		}

		ofs += stride;											// goto next line
	}

	delete [] buf;
	fclose(infile);

	return 1;
}


int floatimage::FromBMP(char *pData, float image_gamma)
{
	InitGamma(image_gamma);
	return FromBMP_Curve(pData, gamma_table, gamma_table, gamma_table);
}


int floatimage::FromBMP_Curve(char *pData, float rcurve[256], float gcurve[256], float bcurve[256])
{
	if (pData == NULL)
		return 0;

	BITMAPINFOHEADER *info = (BITMAPINFOHEADER *) pData;

	Allocate(info->biWidth, info->biHeight);
	if (!line)
	{
		return 0;
	}

	int stride = sizeof(RGB) * sizex;			// stride between lines in bytes
	stride = (stride + 3) & (~0x03);			// rounded up to nearest LONG boundary

	int filestride = (info->biBitCount / 8) * sizex;
	filestride = (filestride + 3) & (~0x03);

	if ((info->biBitCount == 32) || (info->biBitCount == 24))
	{
		int pixelskip = info->biBitCount / 8;

		unsigned char *buf = (unsigned char *) (pData + info->biSize);

		if (info->biCompression == 3L) // BI_BITFIELDS
		{
			buf += 3 * sizeof(DWORD);		// skip bitfield mask
		}

		for (int y = 0; y < sizey; y++)
		{
			unsigned char *cur = buf;
			fRGB *curline = line[sizey - 1 - y];
			for (int x = 0; x < sizex; x++)
			{
				curline[x].r = rcurve[cur[2]];
				curline[x].g = gcurve[cur[1]];
				curline[x].b = bcurve[cur[0]];
				cur += pixelskip;
			}

			buf += filestride;
		}
	}
	else if (info->biBitCount == 8)
	{
		unsigned char *buf = (unsigned char *) (pData + info->biSize); // (unsigned char *) new char[filestride];

		for (int y = 0; y < sizey; y++)
		{
			buf += filestride;

			unsigned char *cur = buf;
			fRGB *curline = line[sizey - 1 - y];
			for (int x = 0; x < sizex; x++)
			{
				curline[x].r = rcurve[*cur];
				curline[x].g = gcurve[*cur];
				curline[x].b = bcurve[*cur];
				cur++;
			}
		}
	}

	return 1;
}


int floatimage::LoadBMP_Curve(const char *filename, float rcurve[256], float gcurve[256], float bcurve[256])
{
	FILE *infile = fopen(filename, "rb");

	if (infile == NULL)
		return 0;

	BITMAPINFOHEADER info;
	if (!ReadBMPHeader(infile, &info))
	{
		fclose(infile);
		return 0;
	}

	Allocate(info.biWidth, info.biHeight);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	int stride = sizeof(RGB) * sizex;			// stride between lines in bytes
	stride = (stride + 3) & (~0x03);			// rounded up to nearest LONG boundary

	int filestride = (info.biBitCount / 8) * sizex;
	filestride = (filestride + 3) & (~0x03);

	if (info.biBitCount == 24)
	{
		RGB *buf = (RGB *) new char[stride];

		for (int y = 0; y < sizey; y++)
		{
			if (fread(buf, filestride, 1, infile) != 1)
			{
				Release();
				delete [] buf;
				fclose(infile);
				return 0;
			}

			RGB *cur = buf;
			fRGB *curline = line[sizey - 1 - y];
			for (int x = 0; x < sizex; x++)
			{
				curline[x].r = rcurve[cur->rgbRed];
				curline[x].g = gcurve[cur->rgbGreen];
				curline[x].b = bcurve[cur->rgbBlue];
				cur++;
			}
		}
		delete buf;
	}
	else if (info.biBitCount == 8)
	{
		unsigned char *buf = (unsigned char *) new char[filestride];

		for (int y = 0; y < sizey; y++)
		{
			if (fread(buf, filestride, 1, infile) != 1)
			{
				Release();
				delete [] buf;
				fclose(infile);
				return 0;
			}

			unsigned char *cur = buf;
			fRGB *curline = line[sizey - 1 - y];
			for (int x = 0; x < sizex; x++)
			{
				curline[x].r = rcurve[*cur];
				curline[x].g = gcurve[*cur];
				curline[x].b = bcurve[*cur];
				cur++;
			}
		}
		delete buf;
	}

	fclose(infile);

	return 1;
}


int floatimage::LoadBMP(const char *filename, float image_gamma)
{
	InitGamma(image_gamma);
	return LoadBMP_Curve(filename, gamma_table, gamma_table, gamma_table);
}


int floatimage::SaveBMP(const char *filename, float image_scale, float image_gamma)
{
	FILE *outfile = fopen(filename, "wb");

	BITMAPINFOHEADER info;
	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = sizex;
	info.biHeight = sizey;
	info.biPlanes = 1;
	info.biBitCount = 24;
	info.biCompression = 0L;
	info.biSizeImage = 0;
	info.biXPelsPerMeter = 2834;
	info.biYPelsPerMeter = 2834;
	info.biClrUsed = 0;
	info.biClrImportant = 0;

	if (!WriteBMPHeader(&info, outfile))
	{
		fclose(outfile);
		return 0;
	}

	unsigned char lookup[65536];
	for (int x = 0; x < 65536; x++)
	{
		float tval = (float) ::pow((double)short2float(x) * image_scale, 1.0 / image_gamma) * 256.0f;
		lookup[x] = (tval > 255.0 ? 255 : (unsigned char) tval);
	}


	for (int y = sizey - 1; y >= 0; y--)
	{
		float *buf = &line[y][0].r;
		for (x = 0; x < sizex; x++)
		{
			if ((fwrite(&lookup[float2short(buf[2])], 1, 1, outfile) != 1) ||
				(fwrite(&lookup[float2short(buf[1])], 1, 1, outfile) != 1) ||
				(fwrite(&lookup[float2short(buf[0])], 1, 1, outfile) != 1))
			{
				fclose(outfile);
				return 0;
			}

			buf += 3;
		}
	}
	
	fclose(outfile);
	return 1;
}

int floatimage::SavePPM(const char *filename, float image_scale, float image_gamma)
{
	return 0;
}


int floatimage::LoadPPM_Curve(const char *filename, float rcurve[256], float gcurve[256], float bcurve[256])
{
	FILE *infile = fopen(filename, "rb");

	if (infile == NULL)
		return 0;

	char temp[256];	

	if ((!fgets(temp, 256, infile)) || (temp[0] != 'P') || (temp[1] != '6'))
	{
		fclose(infile);
		return 0;
	}

	temp[0] = fgetc(infile);
	while (temp[0] == '#')
	{
		while (fgetc(infile) != '\n')
			;
//		fscanf(infile, "\n");
//		readline();
		temp[0] = fgetc(infile);
	}
	fgets(&temp[1], 255, infile);

	int width, height;
	sscanf(temp, "%d %d", &width, &height);

	int colors;
	fscanf(infile, "%d", &colors);
	fgetc(infile);

	if (colors != 255)
	{
		fclose(infile);
		return 0;
	}

	Allocate(width, height);
	if (!line)
	{
		fclose(infile);
		return 0;
	}

	RGB *buf = new RGB[width];

	for (int y = 0; y < sizey; y++)
	{
		if (fread(buf, width * sizeof(RGB), 1, infile) != 1)
		{
			Release();
			delete [] buf;
			fclose(infile);
			return 0;
		}

		RGB *cur = buf;
		fRGB *curline = line[y];
		for (int x = 0; x < sizex; x++)
		{
			curline[x].r = rcurve[cur->rgbBlue];
			curline[x].g = gcurve[cur->rgbGreen];
			curline[x].b = bcurve[cur->rgbRed];
			cur++;
		}
	}

	delete buf;

	fclose(infile);

	return 1;
}


int floatimage::LoadPPM(const char *filename, float image_gamma)
{
	InitGamma(image_gamma);
	return LoadPPM_Curve(filename, gamma_table, gamma_table, gamma_table);
}


void floatimage::SwapData(floatimage &in) 
{
	fRGB **tmp;
	int tmp_x,tmp_y;

	tmp = line; tmp_x = sizex; tmp_y = sizey;
	line = in.line; sizex = in.sizex; sizey = in.sizey;
	in.line = tmp; in.sizex = tmp_x; in.sizey = tmp_y;
}

void floatimage::ResampleFrom(floatimage &in) 
{
	fRGB val;
	float *f = &line[0][0].r;
	float inx,iny;
	int y,x;

	for (y = 0; y < sizey; y++) 
	{
		iny = ((y+0.5f)*in.sizey)/sizey-0.5f;
		for (x = 0; x < sizex; x++) 
		{
			inx = ((x+.5f)*in.sizex)/sizex-0.5f;

			val = in.GetPixelBilinear(inx,iny);

			*f++ = val.r;
			*f++ = val.g;
			*f++ = val.b;
		}
	}
}

void floatimage::GaussianBlur(int variance) 
{
	int i,count;
	float scalefac;
  
	scalefac = 1.f;
	count = 0;
   
	for (i=0;i<variance;i++) 
	{
		blur1x();
		blur1y();
		count++;

		scalefac*=1.0f/16;

		if (count>5 || i==variance-1) 
		{
			scaleimage(scalefac);
			count = 0; scalefac = 1.f;
		}

	}
}

void floatimage::FastGaussianBlur(int variance) 
{

#define m_boxvariance(n) (((2*(n)*(n)*(n)+3*(n)*(n)+(n))/3.f)*(4*(n)+2)/((2*(n)+1)*(2.f*(n)+1)))

	int i,count;
	float variance_so_far,rough_target;
	float scalefac;
	int n;
  
	variance_so_far = 0.f;
	scalefac = 1.f;
	count = 0;

	n = 1;

	while (variance/m_boxvariance(n) > 15 ) n++;
	if (variance/m_boxvariance(n) < 5 ) n--;
	if (n==0) n=1;

	printf("%f %d\n",m_boxvariance(n),n);

	if (variance <= 3) rough_target = 0.f;
	else rough_target = variance - (15 - variance/m_boxvariance(n));

	if (rough_target<0.f) rough_target=0.f;

	printf("target %f\n",rough_target);

	while (n>1) {
		while (variance_so_far + m_boxvariance(n) < rough_target) {
			this->box_blur(n);
			printf("subblur %d\n",n);
			variance_so_far += m_boxvariance(n);
		}
		n = n/2;
	}
   
	for (i=(int)(variance_so_far);i<variance;i++) {
		blur1x();
		blur1y();
		count++;
			printf("clean-up blur\n");

		scalefac*=1.0f/16;

		if (count>5 || i==variance-1) {
			scaleimage(scalefac);
			count = 0; scalefac = 1.f;
		}

	}

#undef m_boxvariance
}

void floatimage::box_blur(int hw) {

	this->boxblurx(hw);
	this->boxblury(hw);
	this->scaleimage(1.f/((2*hw+1)*(2*hw+1)));

}

void floatimage::boxblurx(int hw) {
	int y,x;
	float *on,*left,*right;
	float *f = &line[0][0].r;
	float *saveout,*out;

	saveout = (float*)malloc(sizex*3*sizeof(float));

	for (y = 0; y < sizey; y++) {

		left = f;
		on = f;
		right = f;
		out = saveout;

		out[0] = hw*left[0];
		out[1] = hw*left[1];
		out[2] = hw*left[2];

		for (x=0;x<=hw;x++) {
			out[0] += *right++;
			out[1] += *right++;
			out[2] += *right++;
		}
		out+=3;
		
		for (x=1;x<=hw;x++) {
			*out++ = *(out-3)+(*right++)-left[0];
			*out++ = *(out-3)+(*right++)-left[1];
			*out++ = *(out-3)+(*right++)-left[2];
		}
		for (x=hw+1;x<sizex-hw-1;x++) {
			*out++ = *(out-3)+(*right++)-(*left++);
			*out++ = *(out-3)+(*right++)-(*left++);
			*out++ = *(out-3)+(*right++)-(*left++);
		}
		for (x=sizex-hw-1;x<sizex;x++) {
			*out++ = *(out-3)+right[0]-(*left++);
			*out++ = *(out-3)+right[1]-(*left++);
			*out++ = *(out-3)+right[2]-(*left++);
		}

		memcpy(f,saveout,3*sizex*sizeof(float));

		f+=3*sizex;
	}

	free(saveout);
}

void floatimage::boxblury(int hw) {
	int y,x;
	float *on,*up,*down,*tmp;
	float *f = &line[0][0].r;
	float *saveout,*out;

	saveout = (float*)malloc(sizey*3*sizeof(float));


	for (x = 0; x < sizex; x++) {

		up = f;
		on = f;
		down = f;
		out = saveout;

		out[0] = hw*up[0];
		out[1] = hw*up[1];
		out[2] = hw*up[2];

		for (y=0;y<=hw;y++) {
			out[0] += *down++;
			out[1] += *down++;
			out[2] += *down++;
			down+=3*(sizex-1);
		}
		out+=3;
		
		for (y=1;y<=hw;y++) {
			*out++ = *(out-3)+(*down++)-up[0];
			*out++ = *(out-3)+(*down++)-up[1];
			*out++ = *(out-3)+(*down++)-up[2];
			down+=3*(sizex-1);
		}
		for (y=hw+1;y<sizey-hw-1;y++) {
			*out++ = *(out-3)+(*down++)-(*up++);
			*out++ = *(out-3)+(*down++)-(*up++);
			*out++ = *(out-3)+(*down++)-(*up++);
			down+=3*(sizex-1); 	up+=3*(sizex-1);
		}
		for (y=sizey-hw-1;y<sizey;y++) {
			*out++ = *(out-3)+down[0]-(*up++);
			*out++ = *(out-3)+down[1]-(*up++);
			*out++ = *(out-3)+down[2]-(*up++);
			up+=3*(sizex-1);
		}
		tmp = f;
		for (y=0;y<sizey;y++) {
			*tmp++ = saveout[3*y+0];
			*tmp++ = saveout[3*y+1];
			*tmp++ = saveout[3*y+2];
			tmp+=3*(sizex-1);
		}

		f+=3;
	}

	free(saveout);

}

void floatimage::blur1x() {
  int  y;
  register float *on,*left,*right;
  register float *out;
  register int  x;
  float *f = &line[0][0].r;
  float *saveout;

  saveout = (float*)malloc(sizex*3*sizeof(float));

  for (y = 0; y < sizey; y++) {
  
    out = saveout;
    on = f;
    left = f;
    right = f+3;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;

    left-=3;
    for (x = 1; x < sizex-1; x++) {
		*out++ = *on + *on + *left++ + *right++; on++;
		*out++ = *on + *on + *left++ + *right++; on++;
		*out++ = *on + *on + *left++ + *right++; on++;
    }
	right-=3;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;
    *out++ = *on + *on + *left++ + *right++; on++;

	memcpy(f,saveout,3*sizex*sizeof(float));

	f+=3*sizex;

  }

  	free(saveout);

}

void floatimage::blur1y() {
  int  y;

  register float *on,*up,*down;
  register float *out;
  register int  x;
  float *f = &line[0][0].r;
  float *saveout,*lastout,*tmp;

  saveout = (float*)malloc(sizex*3*sizeof(float));
  lastout = (float*)malloc(sizex*3*sizeof(float));

    out = lastout;
    on = f;
    up = f;
    down = f+3*sizex;
	for (x = 0; x < sizex; x++) {
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
	}

    up-=3*sizex;
	tmp = saveout; saveout = lastout; lastout = tmp;

	for (y = 1; y < sizey-1; y++) {
		out = lastout;
		for (x = 0; x < sizex; x++) {
			*out++ = *on + *on + *up++ + *down++; on++;
			*out++ = *on + *on + *up++ + *down++; on++;
			*out++ = *on + *on + *up++ + *down++; on++;
		}
		memcpy(f,saveout,3*sizex*sizeof(float));
		tmp = saveout; saveout = lastout; lastout = tmp;
		f+=3*sizex;
	}

	down-=3*sizex;
	out = lastout;
	for (x = 0; x < sizex; x++) {
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
		*out++ = *on + *on + *up++ + *down++; on++;
	}

	memcpy(f,saveout,3*sizex*sizeof(float));
	memcpy(f+3*sizex,lastout,3*sizex*sizeof(float));

	free(lastout);
	free(saveout);



}

void floatimage::multiplyimage(float cr,float cg,float cb) 
{
	int  y;
	register int  x;
	float *f = &line[0][0].r;


	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{
			*f++ *= cr;
			*f++ *= cg;
			*f++ *= cb;
		}
	}

}

void floatimage::scaleimage(float s) 
{
	int  y;
	register int  x;
	float *f = &line[0][0].r;

	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{
			*f++ *= s;
			*f++ *= s;
			*f++ *= s;
		}
	}
}

void floatimage::shoulder(float cut) 
{
	float scale,t,invscale,val;
	float *f = &line[0][0].r;
	float *last = &line[sizey-1][sizex-1].b;

	scale=2*(1-cut)/3.14159265f;
	invscale = 1/scale;

	while (f <= last)
	{
		val = *f;
		if (val>cut) 
		{
			t = (val-cut)*invscale;
			*f = scale*(float)atan(t)+cut;
		}
		f++;
	}
}

void floatimage::vignette(float r0, float r1) 
{
	int  y;
	register int  x;
	float radius;
	float reduce;
	float a,t;
	float *f = &line[0][0].r;

	for (y = 0; y < sizey; y++) 
	{
		for (x = 0; x < sizex; x++) 
		{

			radius = (float)sqrt((double)(x-sizex/2)*(x-sizex/2) + (y-sizey/2)*(y-sizey/2));

			if (radius > r0) 
			{
				if (radius < r1) 
				{
					t = 1.0f - (radius-r0)/(r1-r0);
					a = t * 2 - 1;
					reduce = (float)((0.25 * M_PI + 0.5 * asin(a) + 0.5 * a * sqrt( 1 - a*a ))/(0.5 * M_PI));

				} else 
				{
					reduce =0.f;
				}
				*f++ *= reduce;
				*f++ *= reduce;
				*f++ *= reduce;
			}
			else 
			{
				f+=3;
			}
		}
	}
}


void floatimage::Rotate90CW()
{
	floatimage rotated;
	rotated.Allocate(sizey, sizex);

	if (rotated.line)
	{
		for (int y = 0; y < sizey; y++)
			for (int x = 0; x < sizex; x++)
				rotated.line[x][y] = line[y][x];

		fRGB **temp = rotated.line;
		rotated.line = line;
		line = temp;

		sizey = rotated.sizey;
		sizex = rotated.sizex;
	}
	FlipHorizontal();
}


void floatimage::ConvolveNaive(floatimage &kernel)
{
	floatimage tmp;

	tmp.Allocate(sizex, sizey);

	for (int y = 0; y < sizey; y++) 
	{
		for (int x = 0; x < sizex; x++) 
		{
			float	r,g,b;
			float	mr,mg,mb;

			r	=	0;
			g	=	0;
			b	=	0;
			mr	=	0;
			mb	=	0;
			mg	=	0;

			for (int ky = 0; ky < kernel.sizey; ky++) 
			{
				for (int kx = 0; kx < kernel.sizex; kx++)
				{
					int	cx = x + kx - kernel.sizex / 2;
					int	cy = y + ky - kernel.sizey / 2;

					if ((cx < 0) || (cy < 0) || (cx >= sizex) || (cy >= sizey)) 
					{
						mr += kernel.line[ky][kx].r;
						mg += kernel.line[ky][kx].g;
						mb += kernel.line[ky][kx].b;
					}
					else 
					{
						r += kernel.line[ky][kx].r * line[cy][cx].r;
						g += kernel.line[ky][kx].g * line[cy][cx].g;
						b += kernel.line[ky][kx].b * line[cy][cx].b;
					}
				}
			}

			tmp.line[y][x].r = r / (1.0f - mr);
			tmp.line[y][x].g = g / (1.0f - mg);
			tmp.line[y][x].b = b / (1.0f - mb);
		}
	}

	SwapData(tmp);
}


void floatimage::ConvolveNaiveSimple(floatimage &kernel)
{
	floatimage tmp;

	tmp.Allocate(sizex, sizey);

	for (int y = 0; y < sizey; y++) 
	{
		for (int x = 0; x < sizex; x++) 
		{
			float	r,g,b;

			r	=	0;
			g	=	0;
			b	=	0;

			for (int ky = 0; ky < kernel.sizey; ky++) 
			{
				for (int kx = 0; kx < kernel.sizex; kx++)
				{
					int	cx = x + kx - kernel.sizex / 2;
					int	cy = y + ky - kernel.sizey / 2;

					if ((cx < 0) || (cy < 0) || (cx >= sizex) || (cy >= sizey)) 
					{
					}
					else 
					{
						r += kernel.line[ky][kx].r * line[cy][cx].r;
						g += kernel.line[ky][kx].g * line[cy][cx].g;
						b += kernel.line[ky][kx].b * line[cy][cx].b;
					}
				}
			}

			tmp.line[y][x].r = r;
			tmp.line[y][x].g = g;
			tmp.line[y][x].b = b;
		}
	}

	SwapData(tmp);
}


void floatimage::BlitTo(floatimage &dst, int dst_left, int dst_top, int src_left, int src_top, int src_width, int src_height)
{
	if (src_width < 0)
		src_width = sizex + src_width + 1;
	if (src_height < 0)
		src_height = sizey + src_height + 1;

	if (src_left < 0)
	{
		src_width += src_left;
		dst_left -= src_left;
		src_left = 0;
	}
	if (src_top < 0)
	{
		src_height += src_top;
		dst_top -= src_top;
		src_top = 0;
	}
	if (dst_left < 0)
	{
		src_left -= dst_left;
		src_width += dst_left;
		dst_left = 0;
	}
	if (dst_top < 0)
	{
		src_top -= dst_top;
		src_height += dst_top;
		dst_top = 0;
	}
	if (src_left + src_width > sizex)
	{
		src_width = sizex - src_left;
	}
	if (src_top + src_height > sizey)
	{
		src_height = sizey - src_top;
	}
	if (dst_left + src_width > dst.sizex)
	{
		src_width = dst.sizex - dst_left;
	}
	if (dst_top + src_height > dst.sizey)
	{
		src_height = dst.sizey - dst_top;
	}

	if (src_width && src_height)
		for (int y = 0; y < src_height; y++)
		{
			memcpy(&dst.line[dst_top + y][dst_left], &line[src_top + y][src_left], src_width * sizeof(fRGB));
		}
}
