//**************************************************************************
// floatimage.h
//**************************************************************************
// The file contains everything you need to load in a perform simple
// manipulations on .pfm, .ppm, .bmp, or raw files.
// This class can also convert itself from a BMP structure in memory.
// These files include all of the required windows type definitions, so
// they can be compiled either stand alone or as part of a windows app.
//
// floatimage is designed to be used with HDRShop as the basis for a plug-in.
// Written by Chris Tchou (touched up by Jon Cohen)
//**************************************************************************

#ifndef __FLOATIMAGE_H
#define __FLOATIMAGE_H


#include <stdio.h>
#include <math.h>
#include <float.h>
#include <memory.h>

//**************************************************************************
// If this isn't part of a windows project, we need to define windows
// types for compatibility:
//**************************************************************************

#ifndef _WINDOWS_

#define BYTE unsigned char
#define DWORD unsigned long
#define LONG long
#define WORD unsigned short

typedef struct 
{
  BYTE    rgbBlue;
  BYTE    rgbGreen;
  BYTE    rgbRed;
  BYTE    rgbReserved;
} RGBQUAD; 

typedef struct 
{
  WORD    bfType; 
  DWORD   bfSize;
  WORD    bfReserved1;
  WORD    bfReserved2;
  DWORD   bfOffBits;
} BITMAPFILEHEADER;  

typedef struct
{
  DWORD  biSize; 
  LONG   biWidth;
  LONG   biHeight;
  WORD   biPlanes;
  WORD   biBitCount;
  DWORD  biCompression;
  DWORD  biSizeImage;
  LONG   biXPelsPerMeter;
  LONG   biYPelsPerMeter;
  DWORD  biClrUsed; 
  DWORD  biClrImportant;
} BITMAPINFOHEADER;

#endif

//**************************************************************************
// Global variables:
//**************************************************************************

extern float gamma_table[256];  // Gamma lookup table for conversion
extern float cur_gamma;			// current gamma value

//**************************************************************************
// Some utility functions we need
//**************************************************************************

#define FI_MAX(a,b)  ((a) > (b) ? (a) : (b))
#define FI_MIN(a,b)  ((a) < (b) ? (a) : (b))

int ReadBMPHeader(FILE *infile, BITMAPINFOHEADER *i);
int WriteBMPHeader(BITMAPINFOHEADER *i, FILE *outfile);
void CreateGammaTable(float gamma, float *table);
void InitGamma(float gamma = 2.2f);
int RecursiveGCD(int a, int b);


//**************************************************************************
// The datatypes:
//**************************************************************************

struct RGB				// 8 bit per channel color value
{
  BYTE rgbBlue;
  BYTE rgbGreen;
  BYTE rgbRed;
};


struct fRGB				// floating point color value
{
	float r;
	float g;
	float b;

	fRGB()
	{
	}

	fRGB(float r, float g, float b)
	{
		this->r = r;
		this->g = g;
		this->b = b;
	}

	void Set(float r, float g, float b)
	{
		this->r = r;
		this->g = g;
		this->b = b;
	}

	fRGB operator + (fRGB &other)
	{
		return fRGB(r + other.r, g + other.g, b + other.b);
	}

	fRGB operator - (fRGB &other)
	{
		return fRGB(r - other.r, g - other.g, b - other.b);
	}

	fRGB operator / (float div)
	{
		return fRGB(r / div, g / div, b / div);
	}

	fRGB operator * (float mult)
	{
		return fRGB(r * mult, g * mult, b * mult);
	}

	void operator -= (fRGB &other)
	{
		r -= other.r;
		g -= other.g;
		b -= other.b;
	}

	void operator += (fRGB &other)
	{
		r += other.r;
		g += other.g;
		b += other.b;
	}

	void operator *= (float mult)
	{
		r *= mult;
		g *= mult;
		b *= mult;
	}

	int operator != (fRGB other)
	{
		return ((other.r != r) || (other.g != g) || (other.b != b));
	}

	float Intensity()			// standard NTSC intensity
	{
		return 0.299f * r + 0.587f * g + 0.114f * b;
	}
};



class floatimage
{
	public:
		// --------------------- Member variables ------------------------

		fRGB **line;			// 2-d buffer of pixel colors, NULL if not allocated
		int sizex, sizey;		// width and height


		// --------------------- Member functions ------------------------

		void Allocate(int x, int y);		// Call this to allocate the buffer; deallocates any existing image
		void Release();						// Call this to explicitly de-allocate the buffer

		floatimage()
		{
			line = NULL; sizex = 0; sizey = 0;
		}

		floatimage(int width, int height)
		{
			line = NULL; sizex = 0; sizey = 0;
			Allocate(width, height);
		}

		floatimage(const floatimage &in);

		floatimage& operator=(const floatimage &in);

		~floatimage()
		{
			Release();
		}

		int LoadPFM(const char *filename);
		int SavePFM(const char *filename);

		int LoadPPM(const char *filename, float image_gamma = 2.2f);
		int SavePPM(const char *filename, float image_scale = 1.0f, float image_gamma = 2.2f);
		int LoadPPM_Curve(const char *filename, float rcurve[256], float gcurve[256], float bcurve[256]);

		int LoadRAW(const char *filename, int width, int height, int rgb, int byte_reverse, int header = 0);		// load raw binary floating point data
		int SaveRAW(const char *filename);
		int LoadRAW_Bytes(const char *filename, int width, int height, int rgb, float gamma, int header = 0);
		int LoadRAW_Shorts(const char *filename, int width, int height, int rgb, int byte_reverse, int header = 0);
		int LoadRAW_UnsignedShorts(const char *filename, int width, int height, int rgb, int byte_reverse, int header = 0);
		int LoadRAW_Longs(const char *filename, int width, int height, int rgb, int byte_reverse, int header = 0);
		int LoadRAW_UnsignedLongs(const char *filename, int width, int height, int rgb, int byte_reverse, int header = 0);

		int LoadBMP(const char *filename, float image_gamma);
		int SaveBMP(const char *filename, float image_scale, float image_gamma);
		int LoadBMP_Curve(const char *filename, float rcurve[256], float gcurve[256], float bcurve[256]);
		int LoadBMP_SubImage(int startx, int starty, int width, int height, const char *filename, float image_gamma);
		int FromBMP(char *pData, float image_gamma);
		int FromBMP_Curve(char *pData, float rcurve[256], float gcurve[256], float bcurve[256]);

		int LineBytes()		// return the length of an image row, in bytes
		{
			return sizeof(fRGB) * sizex;
		}

		void Clear(float r = 0.0f, float g = 0.0f, float b = 0.0f)
		{
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ = r;
				*f++ = g;
				*f++ = b;
			}
		}

		void HalfSize()		// halve the size of the image in both dimensions, downsampling the image appropriately
		{
			floatimage temp;
			temp.Allocate(sizex / 2, sizey / 2);

			for (int y = 0; y < temp.sizey; y++)
			{
				int cy = 2 * y;
				for (int x = 0; x < temp.sizex; x++)
				{
					int cx = 2 * x;
					temp.line[y][x] = line[cy][cx] + line[cy+1][cx] + line[cy][cx+1] + line[cy+1][cx+1];
				}
			}

			SwapData(temp);
		}

		void DoubleSize()	// double the size of the image in both dimensions, upsampling with point sampling
		{
			floatimage temp;
			temp.Allocate(sizex * 2, sizey * 2);

			for (int y = 0; y < temp.sizey; y++)
			{
				int cy = y / 2;
				for (int x = 0; x < temp.sizex; x++)
				{
					int cx = x / 2;
					temp.line[y][x] = line[cy][cx];
				}
			}

			SwapData(temp);
		}

		void GaussianBlur(int variance);		// apply a gaussian blur - brute force implementation
		void FastGaussianBlur(int variance);	// apply a gaussian blur - fast, good approximation
			void blur1x();							// helper functions for gaussian blur
			void blur1y();
			void box_blur(int hw);
			void boxblurx(int hw);
			void boxblury(int hw);
			void scaleimage(float s);

		void ConvolveNaive(floatimage &kernel);				// brute force convolution
		void ConvolveNaiveSimple(floatimage &kernel);		// brute force convolution, ignore boundaries

		void multiplyimage(float cr,float cg,float cb);		// multiply each channel
		void shoulder(float cut);							// apply a shoulder
		void vignette(float r0, float r1);					// apply a vignette

		void ResampleFrom(floatimage &in);					// bilinear samples "in" to the size of this image
		void SwapData(floatimage &in);						// swaps image buffers with "in"
		void SwapByteOrder();								// swaps high and low order bytes in the floating point data

		void BlitTo(floatimage &dst, int dst_left, int dst_top, int src_left = 0, int src_top = 0, int src_width = -1, int src_height = -1);
		void CropTo(floatimage &dst, int startx, int starty, int width, int height)		{ BlitTo(dst, 0, 0, startx, starty, width, height); }
		void Crop(int startx, int starty, int width, int height)						{ floatimage temp(width, height);  CropTo(temp, startx, starty, width, height); SwapData(temp); }	

		void ClipLower(float min, float replace);			// clip values lower than "min" to "replace"

		void Copy(floatimage &other)						// copy the buffer of other
		{
			Allocate(other.sizex, other.sizey);

			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ = *o++;
				*f++ = *o++;
				*f++ = *o++;
			}
		}

		operator *= (float s)
		{
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ *= s;
				*f++ *= s;
				*f++ *= s;
			}
		}

		floatimage & operator -= (const floatimage &other)
		{
			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ -= *o++;
				*f++ -= *o++;
				*f++ -= *o++;
			}
			return *this;
		}

		floatimage & operator += (const floatimage &other)
		{
			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ += *o++;
				*f++ += *o++;
				*f++ += *o++;
			}
			return *this;
		}

		
		operator *= (floatimage &other)
		{
			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f++ *= *o++;
				*f++ *= *o++;
				*f++ *= *o++;
			}
		}

		operator /= (floatimage &other)
		{
			float *f = &line[0][0].r;
			float *o = &other.line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				if (*o != 0.0f)
					*f /= *o;
				f++; o++;
				if (*o != 0.0f)
					*f /= *o;
				f++; o++;
				if (*o != 0.0f)
					*f /= *o;
				f++; o++;
			}
		}

		void Add(floatimage &a, floatimage &b)
		{
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ + *sb++;
				*d++ = *sa++ + *sb++;
				*d++ = *sa++ + *sb++;
			}			
		}

		void Pow(double p)
		{
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				f[0] = (float) (::pow((double)f[0], p));
				f[1] = (float) (::pow((double)f[1], p));
				f[2] = (float) (::pow((double)f[2], p));
				f += 3;
			}
		}

		void Subtract(floatimage &a, floatimage &b)
		{
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ - *sb++;		// WOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!
				*d++ = *sa++ - *sb++;
				*d++ = *sa++ - *sb++;
			}			
		}

		void Multiply(floatimage &a, floatimage &b)
		{
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ * *sb++;		// WOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!
				*d++ = *sa++ * *sb++;
				*d++ = *sa++ * *sb++;
			}			
		}

		void MultiplyInverse(floatimage &a, floatimage &b)
		{
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				*d++ = *sa++ * (1.0f - *sb++);		// WOOOOOOOOOOOOOOOOOOOOOOOO!!!!!!!!!!!!!
				*d++ = *sa++ * (1.0f - *sb++);
				*d++ = *sa++ * (1.0f - *sb++);
			}			
		}

		void Divide(floatimage &a, floatimage &b)		// warning - not safe
		{
			Allocate(a.sizex, a.sizey);

			float *d = &line[0][0].r;
			float *sa = &a.line[0][0].r;
			float *sb = &b.line[0][0].r;
			float *dlast = &line[sizey-1][sizex-1].b;
			while (d < dlast)
			{
				if (*sb != 0.0f)
					*d = *sa / *sb;
				else
					*d = FLT_MAX;
				d++; sa++; sb++;
			}			
		}

		void AbsoluteValue()
		{
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				*f = (float) fabs(*f); f++;
				*f = (float) fabs(*f); f++;
				*f = (float) fabs(*f); f++;
			}
		}

		void ClampRange(float min, float max)
		{
			float *f = &line[0][0].r;
			float *last = &line[sizey-1][sizex-1].b;
			while (f < last)
			{
				if (*f < min)
					*f = min;
				else if (*f > max)
					*f = max;
				f++;
				if (*f < min)
					*f = min;
				else if (*f > max)
					*f = max;
				f++;
				if (*f < min)
					*f = min;
				else if (*f > max)
					*f = max;
				f++;
			}
		}

		fRGB Maximum(int x0, int y0, int x1, int y1)
		{
			fRGB damax = line[y0][x0];

			for (int y = y0; y <= y1; y++)
			{
				for (int x = x0; x <= x1; x++)
				{
					if (line[y][x].r > damax.r)
						damax.r = line[y][x].r;
					if (line[y][x].g > damax.g)
						damax.g = line[y][x].g;
					if (line[y][x].b > damax.b)
						damax.b = line[y][x].b;
				}
			}

			return damax;
		}


		fRGB Minimum(int x0, int y0, int x1, int y1)
		{
			fRGB damin = line[y0][x0];

			for (int y = y0; y <= y1; y++)
			{
				for (int x = x0; x <= x1; x++)
				{
					if (line[y][x].r < damin.r)
						damin.r = line[y][x].r;
					if (line[y][x].g < damin.g)
						damin.g = line[y][x].g;
					if (line[y][x].b < damin.b)
						damin.b = line[y][x].b;
				}
			}

			return damin;
		}

		void ScaleToRange(float max)
		{
			fRGB maxcolors = Maximum(0, 0, sizex-1, sizey-1);
			float tmax = FI_MAX(FI_MAX(maxcolors.r, maxcolors.g), maxcolors.b);			
			*this *= (max / tmax);
		}

		fRGB Average(int x0, int y0, int x1, int y1)
		{
			int count = 0;
			fRGB dasum(0.0f, 0.0f, 0.0f);

			for (int y = y0; y <= y1; y++)
			{
				for (int x = x0; x <= x1; x++)
				{
					dasum.r += line[y][x].r;
					dasum.g += line[y][x].g;
					dasum.b += line[y][x].b;
					count++;
				}
			}

			dasum.r /= count;
			dasum.g /= count;
			dasum.b /= count;

			return dasum;
		}


		fRGB GetPixelBilinearSlow(float sx, float sy)
		{
			fRGB result;
			int bx = (int) sx;
			int by = (int) sy;
			int bxo = bx + 1;
			int byo = by + 1;
			sx -= bx;
			sy -= by;

			bx %= sizex;
			by %= sizey;
			bxo %= sizex;
			byo %= sizey;

			float osx = 1.0f - sx;
			float osy = 1.0f - sy;

			result.r = line[by][bx].r * osx * osy +
					   line[by][bxo].r * sx * osy +
					   line[byo][bx].r * osx * sy +
					   line[byo][bxo].r * sx * sy;

			result.g = line[by][bx].g * osx * osy +
					   line[by][bxo].g * sx * osy +
					   line[byo][bx].g * osx * sy +
					   line[byo][bxo].g * sx * sy;

			result.b = line[by][bx].b * osx * osy +
					   line[by][bxo].b * sx * osy +
					   line[byo][bx].b * osx * sy +
					   line[byo][bxo].b * sx * sy;

			return result;
		}

		fRGB GetPixelBilinearLogSlow(float sx, float sy)
		{
			fRGB result;

			if (sx < 0.0f)
				sx = 0.0f;
			else if (sx >= sizex - 1.0f)
				sx = sizex - 1.0f;
			if (sy < 0.0f)
				sy = 0.0f;
			else if (sy >= sizey - 1.0f)
				sy = sizey - 1.0f;

			int bx = (int) sx;
			int by = (int) sy;
			int bxo = bx + 1;  if (bxo > sizex - 1)  bxo = sizex - 1;
			int byo = by + 1;  if (byo > sizey - 1)  byo = sizey - 1;

			sx -= bx;
			sy -= by;

			float osx = 1.0f - sx;
			float osy = 1.0f - sy;

			result.r = (float) (exp(log(line[by][bx].r) * osx * osy +
					       log(line[by][bxo].r) * sx * osy +
					       log(line[byo][bx].r) * osx * sy +
					       log(line[byo][bxo].r) * sx * sy));

			result.g = (float) (exp(log(line[by][bx].g) * osx * osy +
					       log(line[by][bxo].g) * sx * osy +
					       log(line[byo][bx].g) * osx * sy +
					       log(line[byo][bxo].g) * sx * sy));

			result.b = (float) (exp(log(line[by][bx].b) * osx * osy +
						   log(line[by][bxo].b) * sx * osy +
					       log(line[byo][bx].b) * osx * sy +
					       log(line[byo][bxo].b) * sx * sy));

			return result;
		}

		fRGB GetPixelBicubicSlow(float sx, float sy)
		{
			float A = -0.75f;
			fRGB val(0, 0, 0);
			float totalweight = 0;

			if (sx < 0.0f)
				sx = 0.0f;
			else if (sx >= sizex - 1.0f)
				sx = sizex - 1.0f;
			if (sy < 0.0f)
				sy = 0.0f;
			else if (sy >= sizey - 1.0f)
				sy = sizey - 1.0f;

			int bx = (int) sx;
			int by = (int) sy;

//			bx %= sizex;
//			by %= sizey;

			for (int y = by - 1; y < by + 3; y++)
			{
				if ((y >= 0) && (y < sizey))
				{
					float disty = (float) (fabs(sy - y));
					float yweight;

					if (disty <= 1)			// photoshop bicubic
						yweight = (( A + 2.0f ) * disty - ( A + 3.0f )) * disty * disty + 1.0f;
					else
						yweight = (( A * disty - 5.0f * A ) * disty + 8.0f * A ) * disty - 4.0f * A;

/*
					if (disty <= 1)			// 'true' bicubic ?
						yweight  = ( ( disty - 9.0/5.0 ) * disty -   1.0/5.0     ) * disty + 1.0;
					else
						yweight  = ( ( -1.0/3.0 * (disty-1) + 4.0/5.0     ) * (disty-1) -   7.0/15.0 ) * (disty-1);
*/
					for (int x = bx - 1; x < bx + 3; x++)
					{
						if ((x >= 0) && (x < sizex))
						{
							float distx = (float) (fabs(sx - x));
							float weight;

							if (distx <= 1)		// photoshop bicubic
								weight = ((( A + 2.0f ) * distx - (A + 3.0f)) * distx * distx +1.0f) * yweight;
							else
								weight = ((( A * distx - 5.0f * A ) * distx + 8.0f * A ) * distx - 4.0f * A) * yweight;

/*							if (distx <= 1)		// 'true' bicubic ?
								weight  = (( ( distx - 9.0/5.0 ) * distx -   1.0/5.0     ) * distx + 1.0) * yweight;
							else
								weight  = (( ( -1.0/3.0 * (distx-1) + 4.0/5.0     ) * (distx-1) -   7.0/15.0 ) * (distx-1)) * yweight;
*/
							val += line[y][x] * weight;
							totalweight += weight;
						}
					}
				}
			}
			val *= 1.0f / totalweight;
			return val;
		}


		fRGB GetPixelBicubicLogSlow(float sx, float sy)
		{
			float A = -0.75f;
			fRGB val(0, 0, 0);
			float totalweight = 0;

			if (sx < 0.0f)
				sx = 0.0f;
			else if (sx >= sizex - 1.0f)
				sx = sizex - 1.0f;
			if (sy < 0.0f)
				sy = 0.0f;
			else if (sy >= sizey - 1.0f)
				sy = sizey - 1.0f;

			int bx = (int) sx;
			int by = (int) sy;

//			bx %= sizex;
//			by %= sizey;

			for (int y = by - 1; y < by + 3; y++)
			{
				if ((y >= 0) && (y < sizey))
				{
					float disty = (float) fabs(sy - y);
					float yweight;

					if (disty <= 1)			// photoshop bicubic
						yweight = (( A + 2.0f ) * disty - ( A + 3.0f )) * disty * disty + 1.0f;
					else
						yweight = (( A * disty - 5.0f * A ) * disty + 8.0f * A ) * disty - 4.0f * A;

/*
					if (disty <= 1)			// 'true' bicubic ?
						yweight  = ( ( disty - 9.0/5.0 ) * disty -   1.0/5.0     ) * disty + 1.0;
					else
						yweight  = ( ( -1.0/3.0 * (disty-1) + 4.0/5.0     ) * (disty-1) -   7.0/15.0 ) * (disty-1);
*/
					for (int x = bx - 1; x < bx + 3; x++)
					{
						if ((x >= 0) && (x < sizex))
						{
							float distx = (float) (fabs(sx - x));
							float weight;

							if (distx <= 1)		// photoshop bicubic
								weight = ((( A + 2.0f ) * distx - (A + 3.0f)) * distx * distx +1.0f) * yweight;
							else
								weight = ((( A * distx - 5.0f * A ) * distx + 8.0f * A ) * distx - 4.0f * A) * yweight;

/*							if (distx <= 1)		// 'true' bicubic ?
								weight  = (( ( distx - 9.0/5.0 ) * distx -   1.0/5.0     ) * distx + 1.0) * yweight;
							else
								weight  = (( ( -1.0/3.0 * (distx-1) + 4.0/5.0     ) * (distx-1) -   7.0/15.0 ) * (distx-1)) * yweight;
*/
							val.r += ((float) log(line[y][x].r)) * weight;
							val.g += ((float) log(line[y][x].g)) * weight;
							val.b += ((float) log(line[y][x].b)) * weight;
//							val += line[y][x] * weight;
							totalweight += weight;
						}
					}
				}
			}
			val *= 1.0f / totalweight;
			val.r = (float) exp(val.r);
			val.g = (float) exp(val.g);
			val.b = (float) exp(val.b);
			return val;
		}

		
		fRGB GetPixelBilinear(float sx, float sy)
		{
			fRGB result;
			int bx = (int) sx;
			int by = (int) sy;
			int bxo = bx + 1;
			int byo = by + 1;
			float w00,w01,w10,w11;
			sx -= bx;
			sy -= by;

			if (bxo>=sizex) bxo=sizex-1;
			if (byo>=sizey) byo=sizey-1;
			if (sx<0.f) bx = bxo = 0;
			if (sy<0.f) by = byo = 0;

			w11 = sx*sy;

			w00 = 1 - sx - sy + w11;
			w10 = sx - w11;
			w01 = sy - w11;

			result.r = line[by][bx].r * w00 +
					   line[by][bxo].r * w10 +
					   line[byo][bx].r * w01 +
					   line[byo][bxo].r * w11;

			result.g = line[by][bx].g * w00 +
					   line[by][bxo].g * w10 +
					   line[byo][bx].g * w01 +
					   line[byo][bxo].g * w11;

			result.b = line[by][bx].b * w00 +
					   line[by][bxo].b * w10 +
					   line[byo][bx].b * w01 +
					   line[byo][bxo].b * w11;

			return result;
		}

		void SwapLines(int a, int b)
		{
			fRGB temp;

			for (int x = 0; x < sizex; x++)
			{	
				temp = line[b][x];
				line[b][x] = line[a][x];
				line[a][x] = temp;
			}
		}

		void SwapColumns(int a, int b)
		{
			fRGB temp;

			for (int y = 0; y < sizey; y++)
			{	
				temp = line[y][b];
				line[y][b] = line[y][a];
				line[y][a] = temp;
			}
		}

		void FlipVertical()
		{
			int top = sizey - 1;
			int bottom = 0;
			while (top > bottom)
			{
				SwapLines(top, bottom);
				top--;
				bottom++;
			}
		}

		void FlipHorizontal()
		{
			int right = sizex - 1;
			int left = 0;
			while (right > left)
			{
				SwapColumns(right, left);
				right--;
				left++;
			}
		}

		void Rotate90CW();


		void GetSubImage(int xstart, int ystart, int xwidth, int yheight, floatimage &dest)
		{
			dest.Allocate(xwidth, yheight);
			for (int y = 0; y < yheight; y++)
				for (int x = 0; x < xwidth; x++)
				{
					dest.line[y][x] = line[ystart + y][xstart + x];
				}
		}



		void ShiftXLeft(int pixels)			// only uses one pixel extra memory, and reads/writes each pixel once
		{									// though it might not be the best in terms of cache coherency...
			pixels %= sizex;
			if (pixels < 0)
				pixels += sizex;
			if (pixels == 0)
				return;

			int GCD = RecursiveGCD(pixels, sizex);

			for (int y = 0; y < sizey; y++)
			{
				for (int p = 0; p < GCD; p++)
				{
					fRGB *liney = line[y];
					fRGB temp = line[y][p];

					int ox = p;
					int x = p + pixels;

					do
					{
						do
						{
							liney[ox] = liney[x];
							ox = x;
							x += pixels;
						} while (x < sizex);

						x -= sizex;
					} while (x != p);

					liney[ox] = temp;
				}
			}
		}


		void ShiftYUp(int pixels)
		{
			pixels %= sizey;
			if (pixels < 0)
				pixels += sizey;
			if (pixels == 0)
				return;

			int GCD = RecursiveGCD(pixels, sizey);

			fRGB *temp = new fRGB[sizex];

			for (int p = 0; p < GCD; p++)
			{
				memcpy(temp, line[p], LineBytes());

				int oy = p;
				int y = p + pixels;

				do
				{
					do
					{
						memcpy(line[oy], line[y], LineBytes());
						oy = y;
						y += pixels;
					} while (y < sizey);

					y -= sizey;
				} while (y != p);

				memcpy(line[oy], temp, LineBytes());
			}

			delete [] temp;
		}

		
		void ShiftX(float percent, floatimage &dest)
		{
			while (percent < 0.0)
				percent += 1.0;
			float shift = percent * sizex;
			dest.Allocate(sizex, sizey);
			for (int y = 0; y < sizey; y++)
				for (int x = 0; x < sizex; x++)
				{
					float sx = x + shift;

					if (sx < 0.0)
						sx += ((int) (sx / sizex) + 1) * sizex;

					sx -= ((int) (sx / sizex)) * sizex;

					dest.line[y][x] = GetPixelBilinearSlow(sx, (float) y);
				}
		}
};



#endif